@@ -25,6 +25,15 @@
 #include "fw.h"
 #include "raw1394_private.h"
 
+/*
+ * ABI version history is documented in linux/firewire-cdev.h.
+ */
+#ifdef FW_CDEV_EVENT_REQUEST2
+#define IMPLEMENTED_CDEV_ABI_VERSION	4
+#else
+#define IMPLEMENTED_CDEV_ABI_VERSION	2
+#endif
+
 int
 fw_errcode_to_errno(raw1394_errcode_t errcode)
 {
@@ -150,7 +159,7 @@ scan_devices(fw_handle_t handle)
 			continue;
 		memset(&get_info, 0, sizeof(get_info));
 		memset(&reset, 0, sizeof(reset));
-		get_info.version = FW_CDEV_VERSION;
+		get_info.version = IMPLEMENTED_CDEV_ABI_VERSION;
 		get_info.rom = 0;
 		get_info.rom_length = 0;
 		get_info.bus_reset = ptr_to_u64(&reset);
@@ -210,41 +219,44 @@ handle_lost_device(fw_handle_t handle, int i)
 
 struct address_closure {
 	int (*callback)(raw1394handle_t handle, struct address_closure *ac,
-			struct fw_cdev_event_request *request, int i);
+			int tcode, unsigned long long offset,
+			int source_node_id, int card, unsigned kernel_handle,
+			size_t length, void *data);
 };
 
 static int
 handle_fcp_request(raw1394handle_t handle, struct address_closure *ac,
-		   struct fw_cdev_event_request *request, int i)
+		   int tcode, unsigned long long offset, int source_node_id,
+		   int card, unsigned kernel_handle, size_t length, void *data)
 {
 	struct fw_cdev_send_response response;
 	int is_response;
 
-	response.handle = request->handle;
-	response.rcode  = RCODE_COMPLETE;
-	response.length = 0;
-	response.data   = 0;
+	response.handle	= kernel_handle;
+	response.rcode	= RCODE_COMPLETE;
+	response.length	= 0;
+	response.data	= 0;
 
 	if (handle->mode.fw->fcp_handler == NULL)
 		response.rcode = RCODE_ADDRESS_ERROR;
 
-	if (request->tcode >= TCODE_WRITE_RESPONSE)
+	if (tcode >= TCODE_WRITE_RESPONSE)
 		response.rcode = RCODE_CONFLICT_ERROR;
 
-	if (ioctl(handle->mode.fw->devices[i].fd,
+	if (ioctl(handle->mode.fw->ioctl_fd,
 		  FW_CDEV_IOC_SEND_RESPONSE, &response) < 0)
 		return -1;
 
+	if (card != handle->mode.fw->card)
+		return 0;
+
 	if (response.rcode != RCODE_COMPLETE)
 		return 0;
 
-	is_response = request->offset >= CSR_REGISTER_BASE + CSR_FCP_RESPONSE;
+	is_response = offset >= CSR_REGISTER_BASE + CSR_FCP_RESPONSE;
 
-	return handle->mode.fw->fcp_handler(handle,
-				   handle->mode.fw->devices[i].node_id,
-				   is_response,
-				   request->length,
-				   (unsigned char *) request->data);
+	return handle->mode.fw->fcp_handler(handle, source_node_id,
+					    is_response, length, data);
 }
 
 static int
@@ -262,7 +274,7 @@ handle_device_event(raw1394handle_t handle,
 	int i;
 
 	i = device - fwhandle->devices;
-	if (events == EPOLLHUP)
+	if (events & EPOLLHUP)
 		return handle_lost_device(fwhandle, i);
 
 	len = read(fwhandle->devices[i].fd,
@@ -291,8 +303,9 @@ handle_device_event(raw1394handle_t handle,
 	case FW_CDEV_EVENT_RESPONSE:
 		rc = u64_to_ptr(u->response.closure);
 
-		if (rc->data != NULL)
-			memcpy(rc->data, u->response.data, rc->length);
+		/* Kernel ensures that u->response.length does not overflow. */
+		if (rc->data != NULL && u->response.rcode == RCODE_COMPLETE)
+			memcpy(rc->data, u->response.data, u->response.length);
 
 		errcode = fw_to_raw1394_errcode(u->response.rcode);
 		tag = rc->tag;
@@ -302,7 +315,24 @@ handle_device_event(raw1394handle_t handle,
 
 	case FW_CDEV_EVENT_REQUEST:
 		ac = u64_to_ptr(u->request.closure);
-		return ac->callback(handle, ac, &u->request, i);
+		return ac->callback(handle, ac, u->request.tcode,
+				    u->request.offset,
+				    /* wild guess, but can't do better */
+				    fwhandle->devices[i].node_id,
+				    fwhandle->card,
+				    u->request.handle,
+				    u->request.length, u->request.data);
+
+#ifdef FW_CDEV_EVENT_REQUEST2 /* added in kernel 2.6.36 */
+	case FW_CDEV_EVENT_REQUEST2:
+		ac = u64_to_ptr(u->request.closure);
+		return ac->callback(handle, ac, u->request2.tcode,
+				    u->request2.offset,
+				    u->request2.source_node_id,
+				    u->request2.card,
+				    u->request2.handle,
+				    u->request2.length, u->request2.data);
+#endif
 
 #ifdef FW_CDEV_EVENT_ISO_RESOURCE_ALLOCATED /* added in kernel 2.6.30 */
 	case FW_CDEV_EVENT_ISO_RESOURCE_ALLOCATED:
@@ -311,6 +341,16 @@ handle_device_event(raw1394handle_t handle,
 		       sizeof u->iso_resource);
 		return 0;
 #endif
+
+#ifdef FW_CDEV_EVENT_PHY_PACKET_SENT /* added in kernel 2.6.36 */
+	case FW_CDEV_EVENT_PHY_PACKET_SENT:
+		rc = u64_to_ptr(u->phy_packet.closure);
+		errcode = fw_to_raw1394_errcode(u->phy_packet.rcode);
+		tag = rc->tag;
+		free(rc);
+
+		return fwhandle->tag_handler(handle, tag, errcode);
+#endif
 	default:
 	case FW_CDEV_EVENT_ISO_INTERRUPT:
 		/* Never happens. */
@@ -355,7 +395,7 @@ handle_inotify(raw1394handle_t handle, struct epoll_closure *ec,
 		}
 	}
 
-	info.version = FW_CDEV_VERSION;
+	info.version = IMPLEMENTED_CDEV_ABI_VERSION;
 	info.rom = 0;
 	info.rom_length = 0;
 	info.bus_reset = ptr_to_u64(&reset);
@@ -419,9 +459,13 @@ fw_handle_t fw_new_handle(void)
 	struct epoll_event ep;
 	int i;
 
-	memset(&ep, 0, sizeof(ep));
-
 	handle = malloc(sizeof *handle);
+	if (handle == NULL) {
+		errno = ENOMEM;
+		return NULL;
+	}
+
+	memset(&ep, 0, sizeof(ep));
 	memset(handle, 0, sizeof(*handle));
 
 	handle->tag_handler = default_tag_handler;
@@ -498,9 +542,8 @@ void fw_destroy_handle(fw_handle_t handle)
 
 		close(handle->devices[i].fd);
 	}
-
+	fw_iso_shutdown(handle);
 	close(handle->epoll_fd);
-
 	free(handle);
 
 	return;
@@ -604,7 +647,7 @@ int fw_set_port(fw_handle_t handle, int port)
 
 		memset(&get_info, 0, sizeof(get_info));
 		memset(&reset, 0, sizeof(reset));
-		get_info.version = FW_CDEV_VERSION;
+		get_info.version = IMPLEMENTED_CDEV_ABI_VERSION;
 		get_info.rom = 0;
 		get_info.rom_length = 0;
 		get_info.bus_reset = ptr_to_u64(&reset);
@@ -646,6 +689,7 @@ int fw_set_port(fw_handle_t handle, int port)
 		if (reset.node_id == reset.local_node_id)
 			handle->local_device = &handle->devices[i];
 
+		handle->card = get_info.card;
 		handle->generation = reset.generation;
 		handle->abi_version = get_info.version;
 
@@ -687,7 +731,7 @@ struct request_response_block {
 };
 
 struct allocation {
-	struct address_closure closure;
+	struct address_closure closure; /* must be first member */
 	struct allocation *next;
 	__u32 handle;
 	byte_t *buffer;
@@ -702,7 +746,8 @@ struct allocation {
 
 static int
 handle_arm_request(raw1394handle_t handle, struct address_closure *ac,
-		   struct fw_cdev_event_request *request, int i)
+		   int tcode, unsigned long long offset, int source_node_id,
+		   int card, unsigned kernel_handle, size_t length, void *data)
 {
 	fw_handle_t fwhandle = handle->mode.fw;
 	struct allocation *allocation = (struct allocation *) ac;
@@ -710,19 +755,16 @@ handle_arm_request(raw1394handle_t handle, struct address_closure *ac,
 	struct fw_cdev_send_response response;
 	arm_options_t type;
 	size_t in_length;
-	int offset;
+	int pos;
 
-	offset = request->offset - allocation->offset;
-	response.handle = request->handle;
+	pos = offset - allocation->offset;
+	response.handle = kernel_handle;
 
-	switch (request->tcode) {
+	switch (tcode) {
 	case TCODE_WRITE_QUADLET_REQUEST:
 	case TCODE_WRITE_BLOCK_REQUEST:
-		printf("got write request, offset=0x%012llx, length=%d\n",
-		       request->offset, request->length);
-
 		type = RAW1394_ARM_WRITE;
-		in_length = request->length;
+		in_length = length;
 		response.rcode  = RCODE_COMPLETE;
 		response.length = 0;
 		response.data   = 0;
@@ -730,19 +772,27 @@ handle_arm_request(raw1394handle_t handle, struct address_closure *ac,
 
 	case TCODE_READ_QUADLET_REQUEST:
 	case TCODE_READ_BLOCK_REQUEST:
-		printf("got read request, offset=0x%012llx, length=%d\n",
-		       request->offset, request->length);
-
 		type = RAW1394_ARM_READ;
 		in_length = 0;
 		response.rcode = RCODE_COMPLETE;
-		response.length = request->length;
-		response.data = ptr_to_u64(allocation->data + offset);
+		response.length = length;
+		response.data = ptr_to_u64(allocation->data + pos);
 		break;
 
 	case TCODE_LOCK_REQUEST:
+		/*
+		 * TCODE_LOCK_REQUEST is generated by ABI v3 and older, cannot
+		 * be handled.  Fall through for now.  FIXME.
+		 */
+	case TCODE_LOCK_MASK_SWAP:
+	case TCODE_LOCK_COMPARE_SWAP:
+	case TCODE_LOCK_FETCH_ADD:
+	case TCODE_LOCK_LITTLE_ADD:
+	case TCODE_LOCK_BOUNDED_ADD:
+	case TCODE_LOCK_WRAP_ADD:
+	case TCODE_LOCK_VENDOR_DEPENDENT:
 		type = RAW1394_ARM_LOCK;
-		in_length = request->length;
+		in_length = length;
 		response.length = 4;
 		break;
 
@@ -756,54 +806,62 @@ handle_arm_request(raw1394handle_t handle, struct address_closure *ac,
 		response.rcode  = RCODE_TYPE_ERROR;
 		response.length = 0;
 		response.data   = 0;
-		if (ioctl(fwhandle->devices[i].fd,
+		if (ioctl(fwhandle->ioctl_fd,
 			  FW_CDEV_IOC_SEND_RESPONSE, &response) < 0)
 			return -1;
 	} else if (!(allocation->client_transactions & type)) {
 		if (type == RAW1394_ARM_WRITE)
-			memcpy(allocation->data + offset,
-			       request->data, request->length);
+			memcpy(allocation->data + pos, data, length);
 		else if (type == RAW1394_ARM_LOCK)
 			/* FIXME: do lock ops here */;
 
-		if (ioctl(fwhandle->devices[i].fd,
+		if (ioctl(fwhandle->ioctl_fd,
 			  FW_CDEV_IOC_SEND_RESPONSE, &response) < 0)
 			return -1;
 	}
 
+	/*
+	 * libraw1394 clients do not expect requests from nodes on
+	 * a card other than the one set by raw1394_set_port().
+	 */
+	if (card != fwhandle->card)
+		return 0;
+
 	if (!(allocation->notification_options & type))
 		return 0;
 
 	rrb = malloc(sizeof *rrb + in_length + response.length);
+	if (rrb == NULL) {
+		errno == ENOMEM;
+		return -1;
+	}
 
 	rrb->request_response.request = &rrb->request;
 	rrb->request_response.response = &rrb->response;
 
 	rrb->request.destination_nodeid = fwhandle->reset.local_node_id;
-	rrb->request.source_nodeid = fwhandle->devices[i].node_id;
-	rrb->request.destination_offset = request->offset;
+	rrb->request.source_nodeid      = source_node_id;
+	rrb->request.destination_offset = offset;
 	rrb->request.tlabel = 0;
-	if (request->tcode < 0x10) {
-		rrb->request.tcode = request->tcode;
+	if (tcode < 0x10) {
+		rrb->request.tcode = tcode;
 		rrb->request.extended_transaction_code = 0;
 	} else {
 		rrb->request.tcode = TCODE_LOCK_REQUEST;
-		rrb->request.extended_transaction_code = request->tcode - 0x10;
+		rrb->request.extended_transaction_code = tcode - 0x10;
 	}
 	rrb->request.generation = fwhandle->reset.generation;
 	rrb->request.buffer_length = in_length;
 	rrb->request.buffer = rrb->data;
-	memcpy(rrb->request.buffer, request->data, in_length);
+	memcpy(rrb->request.buffer, data, in_length);
 
 	rrb->response.response_code = response.rcode;
 	rrb->response.buffer_length = response.length;
 	rrb->response.buffer = rrb->data + in_length;
-	memcpy(rrb->response.buffer,
-	       allocation->data + offset, response.length);
+	memcpy(rrb->response.buffer, allocation->data + pos, response.length);
 
 	return fwhandle->arm_tag_handler(handle, allocation->tag, type,
-				       request->length,
-				       &rrb->request_response);
+					 length, &rrb->request_response);
 }
 
 int
@@ -818,8 +876,10 @@ fw_arm_register(fw_handle_t handle, nodeaddr_t start,
 	int retval;
 
 	allocation = malloc(sizeof *allocation + length);
-	if (allocation == NULL)
+	if (allocation == NULL) {
+		errno = ENOMEM;
 		return -1;
+	}
 
 	allocation->closure.callback = handle_arm_request;
 	allocation->buffer = initial_value;
@@ -835,7 +895,9 @@ fw_arm_register(fw_handle_t handle, nodeaddr_t start,
 	request.offset = start;
 	request.length = length;
 	request.closure = ptr_to_u64(&allocation->closure);
-
+#if IMPLEMENTED_CDEV_ABI_VERSION >= 4
+	request.region_end = start + length;
+#endif
 	retval = ioctl(handle->ioctl_fd, FW_CDEV_IOC_ALLOCATE, &request);
 	if (retval < 0) {
 		free(allocation);
@@ -930,20 +992,6 @@ int fw_wake_up(fw_handle_t handle)
 	return fw_echo_request(handle, 0);
 }
 
-int fw_phy_packet_write (fw_handle_t handle, quadlet_t data)
-{
-	errno = ENOSYS;
-	return -1;
-}
-
-int
-fw_start_phy_packet_write(fw_handle_t handle,
-			       quadlet_t data, unsigned long tag)
-{
-	errno = ENOSYS;
-	return -1;
-}
-
 static int
 send_request(fw_handle_t handle, int tcode,
 	     nodeid_t node, nodeaddr_t addr,
@@ -1005,9 +1053,8 @@ node_id_ok:
 		return -1;
 	}
 
-	closure->data   = out;
-	closure->length = out_length;
-	closure->tag    = tag;
+	closure->data = out;
+	closure->tag  = tag;
 
 	request = (struct fw_cdev_send_request *) handle->buffer;
 	request->tcode      = tcode;
@@ -1165,6 +1212,46 @@ fw_start_async_stream(fw_handle_t handle, unsigned int channel,
 			    0, addr, length, data, 0, NULL, rawtag);
 }
 
+int
+fw_start_phy_packet_write(fw_handle_t handle, quadlet_t data, unsigned long tag)
+{
+#ifdef FW_CDEV_IOC_SEND_PHY_PACKET /* added in kernel 2.6.36 */
+	struct fw_cdev_send_phy_packet send_phy_packet;
+	struct request_closure *closure;
+	int retval;
+
+	if (handle->local_device == NULL) {
+		handle->err = -EPERM;
+		errno = EPERM;
+		return -1;
+	}
+
+	closure = malloc(sizeof *closure);
+	if (closure == NULL) {
+		handle->err = -RCODE_SEND_ERROR;
+		errno = fw_errcode_to_errno(handle->err);
+		return -1;
+	}
+
+	closure->data = NULL;
+	closure->tag  = tag;
+
+	send_phy_packet.closure    = ptr_to_u64(closure);
+	send_phy_packet.data[0]    =  be32_to_cpu(data);
+	send_phy_packet.data[1]    = ~be32_to_cpu(data);
+	send_phy_packet.generation = handle->local_device->generation;
+	retval = ioctl(handle->local_device->fd, FW_CDEV_IOC_SEND_PHY_PACKET,
+		       &send_phy_packet);
+	if (retval < 0)
+		free(closure);
+
+	return retval;
+#else
+	errno = ENOSYS;
+	handle->err = -errno;
+	return -1;
+#endif
+}
 
 int
 fw_start_async_send(fw_handle_t handle,
@@ -1300,6 +1387,32 @@ fw_async_stream(raw1394handle_t handle, unsigned int channel,
 				 0, addr, length, data, 0, NULL);
 }
 
+int
+fw_phy_packet_write(raw1394handle_t handle, quadlet_t data)
+{
+	fw_handle_t fwhandle = handle->mode.fw;
+	struct raw1394_reqhandle reqhandle;
+	struct sync_data sd = { 0, 0 };
+	int err;
+
+	reqhandle.callback = sync_callback;
+	reqhandle.data = &sd;
+
+	err = fw_start_phy_packet_write(fwhandle, data,
+					(unsigned long) &reqhandle);
+
+	while (!sd.done) {
+		if (err < 0)
+			return err;
+		err = fw_loop_iterate(handle);
+	}
+
+	fwhandle->err = sd.err;
+	errno = fw_errcode_to_errno(sd.err);
+
+	return (errno ? -1 : 0);
+}
+
 int
 fw_async_send(fw_handle_t handle,
 		   size_t length, size_t header_length,
@@ -1318,8 +1431,10 @@ fw_start_fcp_listen(fw_handle_t handle)
 	struct address_closure *closure;
 
 	closure = malloc(sizeof *closure);
-	if (closure == NULL)
+	if (closure == NULL) {
+		errno = ENOMEM;
 		return -1;
+	}
 
 	closure->callback = handle_fcp_request;
 
@@ -1327,6 +1442,9 @@ fw_start_fcp_listen(fw_handle_t handle)
 	request.offset = CSR_REGISTER_BASE + CSR_FCP_COMMAND;
 	request.length = CSR_FCP_END - CSR_FCP_COMMAND;
 	request.closure = ptr_to_u64(closure);
+#if IMPLEMENTED_CDEV_ABI_VERSION >= 4
+	request.region_end = CSR_REGISTER_BASE + CSR_FCP_END;
+#endif
 	if (ioctl(handle->ioctl_fd, FW_CDEV_IOC_ALLOCATE, &request) < 0)
 		return -1;
 
@@ -1367,7 +1485,7 @@ fw_get_config_rom(fw_handle_t handle, quadlet_t *buffer,
 	}
 
 	memset(&get_info, 0, sizeof(get_info));
-	get_info.version = FW_CDEV_VERSION;
+	get_info.version = IMPLEMENTED_CDEV_ABI_VERSION;
 	get_info.rom = ptr_to_u64(buffer);
 	get_info.rom_length = buffersize;
 	get_info.bus_reset = 0;
@@ -1406,6 +1524,7 @@ iso_resource_modify(raw1394handle_t handle, unsigned int bandwidth,
 	resource.closure = ptr_to_u64(&event);
 	resource.channels = channel >= 0 ? 1ULL << channel : 0;
 	resource.bandwidth = bandwidth;
+	resource.handle = 0; /* unused */
 
 	ioctl_nr = mode == RAW1394_MODIFY_ALLOC ?
 		   FW_CDEV_IOC_ALLOCATE_ISO_RESOURCE_ONCE :
@@ -1428,7 +1547,7 @@ iso_resource_modify(raw1394handle_t handle, unsigned int bandwidth,
 	return 0;
 }
 
-static inline int abi_v2_available(raw1394handle_t handle)
+static inline int abi_has_iso_resource_management(raw1394handle_t handle)
 {
 	return handle->mode.fw->abi_version >= 2;
 }
@@ -1442,7 +1561,7 @@ iso_resource_modify(raw1394handle_t handle, unsigned int bandwidth,
 	return -1;
 }
 
-static inline int abi_v2_available(raw1394handle_t handle)
+static inline int abi_has_iso_resource_management(raw1394handle_t handle)
 {
 	return 0;
 }
@@ -1453,7 +1572,7 @@ int
 fw_bandwidth_modify(raw1394handle_t handle, unsigned int bandwidth,
 		    enum raw1394_modify_mode mode)
 {
-	if (abi_v2_available(handle))
+	if (abi_has_iso_resource_management(handle))
 		return iso_resource_modify(handle, bandwidth, -1, mode);
 	else
 		return ieee1394_bandwidth_modify(handle, bandwidth, mode);
@@ -1463,7 +1582,7 @@ int
 fw_channel_modify(raw1394handle_t handle, unsigned int channel,
 		  enum raw1394_modify_mode mode)
 {
-	if (abi_v2_available(handle))
+	if (abi_has_iso_resource_management(handle))
 		return iso_resource_modify(handle, 0, channel, mode);
 	else
 		return ieee1394_channel_modify(handle, channel, mode);
